from __future__ import annotations

import os
import io
import pathlib
import datetime
import zipfile
from typing import Iterator

import rich.console
import rich.panel
import pytest
import typer.testing
import click.testing
from pytest_httpx import HTTPXMock
from pytest_golden.plugin import GoldenTestFixture  # type: ignore[import]
from time_machine import TimeMachineFixture

from taas_client import cli, __version__

FROZEN_TIME = datetime.datetime(2022, 10, 12, 13, 37).astimezone()
GOLDEN_PATH = pathlib.Path("golden")


class ExpectedConsole:

    """Object to store expected prints.

    This is used to make testing rich output easier."""

    def __init__(self) -> None:
        self._file = io.StringIO()
        self._console = rich.console.Console(file=self._file)
        self.checked = False
        self.exit_code: int | None = None

    def check(self, result: click.testing.Result) -> None:
        self.checked = True
        assert result.exit_code == self.exit_code
        assert result.output == self._file.getvalue()

    def error_panel(self, text: str) -> None:
        self._console.print(rich.panel.Panel(text, title="Error", title_align="left"))

    def line(self, text: str) -> None:
        self._console.print(text)

    def usage_lines(self) -> None:
        self._console.print("Usage: upload-file [OPTIONS] NOTEBOOK [URL]")
        self._console.print("Try 'upload-file --help' for help.")


class Runner:

    """Thin wrapper around the click test runner."""

    def __init__(self) -> None:
        self._runner = typer.testing.CliRunner()

    def run(self, *args: str | pathlib.Path) -> click.testing.Result:
        result = self._runner.invoke(
            cli.app,
            [str(a) for a in args],
            catch_exceptions=False,
        )
        print(result.output)
        return result


@pytest.fixture
def expected() -> Iterator[ExpectedConsole]:
    console = ExpectedConsole()
    yield console
    assert console.checked


@pytest.fixture
def runner() -> Runner:
    return Runner()


class TestArgumentParsing:
    def test_no_args(self, runner: Runner, expected: ExpectedConsole) -> None:
        result = runner.run()

        expected.usage_lines()
        expected.error_panel("Missing argument 'NOTEBOOK'.")
        expected.exit_code = 2

        expected.check(result)

    def test_inexistent_file(
        self, runner: Runner, tmp_path: pathlib.Path, expected: ExpectedConsole
    ) -> None:
        nb_path = tmp_path / "nb.ipynb"
        result = runner.run(nb_path)

        msg = f"Invalid value for 'NOTEBOOK': File '{nb_path}' does not exist."
        expected.usage_lines()
        expected.error_panel(msg)
        expected.exit_code = 2

        expected.check(result)

    def test_dir(
        self, runner: Runner, tmp_path: pathlib.Path, expected: ExpectedConsole
    ) -> None:
        nb_path = tmp_path / "nb.ipynb"
        nb_path.mkdir()

        result = runner.run(nb_path)

        msg = f"Invalid value for 'NOTEBOOK': File '{nb_path}' is a directory."
        expected.usage_lines()
        expected.error_panel(msg)
        expected.exit_code = 2

        expected.check(result)

    def test_unreadable(
        self, runner: Runner, tmp_path: pathlib.Path, expected: ExpectedConsole
    ) -> None:
        nb_path = tmp_path / "nb.ipynb"
        nb_path.touch()
        nb_path.chmod(0)

        if os.access(nb_path, os.R_OK):  # pragma: no cover
            # doesn't work on CI due to Docker weirdness
            pytest.skip("failed to make file unreadable")
        else:  # pragma: no cover
            result = runner.run(nb_path)

            msg = f"Invalid value for 'NOTEBOOK': File '{nb_path}' is not readable."
            expected.usage_lines()
            expected.error_panel(msg)
            expected.exit_code = 2

            expected.check(result)


def test_wrong_location(
    runner: Runner,
    tmp_path: pathlib.Path,
    expected: ExpectedConsole,
    time_machine: TimeMachineFixture,
) -> None:
    time_machine.move_to(FROZEN_TIME)

    nb_path = tmp_path / "nb.ipynb"
    nb_path.touch()
    result = runner.run(nb_path)

    expected.line("Refusing to submit file from unexpected location: ")
    expected.line(str(nb_path))
    expected.line(
        "Make sure you're not accidentally working in the 'Originale' folder!"
    )
    expected.line("💥 Submission failed! (2022-10-12 13:37:00)")
    expected.exit_code = 1

    expected.check(result)


def test_wrong_location_force(
    runner: Runner,
    tmp_path: pathlib.Path,
    httpx_mock: HTTPXMock,
    golden: GoldenTestFixture,
    expected: ExpectedConsole,
    time_machine: TimeMachineFixture,
) -> None:
    nb_path = tmp_path / "nb.ipynb"
    nb_path.touch()

    time_machine.move_to(FROZEN_TIME)
    ts = FROZEN_TIME.timestamp()
    os.utime(nb_path, (ts, ts))  # atime, mtime

    golden = golden.open(GOLDEN_PATH / "simple.yml")

    httpx_mock.add_response(
        url="http://localhost:5000/testbook",
        method="POST",
        json=golden["input"],
    )

    result = runner.run("--force", nb_path)

    expected.line("Refusing to submit file from unexpected location: ")
    expected.line(str(nb_path))
    expected.line(
        "Make sure you're not accidentally working in the 'Originale' folder!"
    )
    expected.line("Last change: 0 seconds ago")
    expected.line("📫 ✔ Submission successful! (2022-10-12 13:37:00, abcdef)")
    expected.exit_code = 0

    expected.check(result)


@pytest.fixture
def nb_arg(tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch) -> pathlib.Path:
    """Set up a valid notebook at the proper location."""
    monkeypatch.setenv("HOME", str(tmp_path))  # POSIX
    monkeypatch.setenv("USERPROFILE", str(tmp_path))  # Windows

    lab_dir = tmp_path / "work" / "AutPy" / "1-getting-started" / "01-python-basic"
    lab_dir.mkdir(parents=True)
    lab_path = lab_dir / "python-basics.ipynb"
    lab_path.touch()

    ts = FROZEN_TIME.timestamp()
    os.utime(lab_path, (ts, ts))  # atime, mtime

    return lab_path


@pytest.mark.golden_test(str(GOLDEN_PATH / "*.yml"))
def test_golden(
    runner: Runner,
    golden: GoldenTestFixture,
    httpx_mock: HTTPXMock,
    time_machine: TimeMachineFixture,
    nb_arg: pathlib.Path,
) -> None:
    httpx_mock.add_response(
        url="http://localhost:5000/testbook",
        method="POST",
        json=golden["input"],
        status_code=golden.get("http_status", 200),
    )

    delay = golden.get("delay", 0)
    time_machine.move_to(FROZEN_TIME + datetime.timedelta(seconds=delay))

    result = runner.run(nb_arg)

    exit_code = golden.get("exit_code", 0)
    assert result.exit_code == exit_code
    assert result.stdout == golden.out["output"]


@pytest.fixture
def golden_special(
    request: pytest.FixtureRequest, golden: GoldenTestFixture
) -> GoldenTestFixture:
    name = request.node.name[len("test_") :]
    name = name.replace("[", "_").replace("]", "")
    return golden.open(GOLDEN_PATH / "special" / f"{name}.yml")


PROJECT_URL = "http://localhost:5000/project"


@pytest.fixture
def project_notebook(tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch) -> None:
    cwd_path = tmp_path / "cwd"
    cwd_path.mkdir()
    monkeypatch.chdir(cwd_path)

    nb_path = cwd_path / "project.ipynb"
    nb_path.touch()


def test_project(
    runner: Runner,
    golden_special: GoldenTestFixture,
    httpx_mock: HTTPXMock,
    time_machine: TimeMachineFixture,
    tmp_path: pathlib.Path,
    project_notebook: None,
) -> None:
    httpx_mock.add_response(
        url=PROJECT_URL, method="POST", json=golden_special["input"]
    )
    time_machine.move_to(FROZEN_TIME)

    project_path = tmp_path / "project.zip"
    with zipfile.ZipFile(project_path, "w"):
        pass

    result = runner.run("--project", str(project_path))

    assert result.exit_code == 0
    assert result.stdout == golden_special.out["output"]

    request = httpx_mock.get_request()
    assert request is not None
    lines = request.read().split(b"\r\n")
    expected = [
        b'Content-Disposition: form-data; name="file"; filename="project.zip"',
        b'Content-Disposition: form-data; name="notebook"; filename="project.ipynb"',
    ]
    for line in expected:
        assert line in lines


@pytest.mark.parametrize(
    "hard_limit, force",
    [
        pytest.param(False, False, id="soft"),
        pytest.param(False, True, id="soft_force"),
        pytest.param(True, False, id="hard"),
        pytest.param(True, True, id="hard_force"),
    ],
)
def test_project_size_limit(
    runner: Runner,
    golden_special: GoldenTestFixture,
    httpx_mock: HTTPXMock,
    time_machine: TimeMachineFixture,
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    hard_limit: bool,
    force: bool,
    project_notebook: None,
) -> None:
    if force and not hard_limit:
        httpx_mock.add_response(
            url=PROJECT_URL, method="POST", json=golden_special["input"]
        )

    time_machine.move_to(FROZEN_TIME)
    monkeypatch.setattr(cli, "PROJECT_LIMIT_SOFT_MB", 1)
    monkeypatch.setattr(cli, "PROJECT_LIMIT_HARD_MB", 2)

    file_size_mb = 2 if hard_limit else 1
    project_path = tmp_path / "project.zip"
    with zipfile.ZipFile(project_path, "w") as zf:
        zf.writestr("data.bin", "\x00" * file_size_mb * 1024 * 1024)

    flags = ["--project"]
    if force:
        flags.append("--force")
    result = runner.run(*flags, str(project_path))

    assert result.exit_code == golden_special.out["exit_code"]
    assert result.stdout == golden_special.out["output"]


def test_project_invalid_zip(
    runner: Runner,
    tmp_path: pathlib.Path,
    time_machine: TimeMachineFixture,
    golden_special: GoldenTestFixture,
) -> None:
    time_machine.move_to(FROZEN_TIME)
    project_path = tmp_path / "project.zip"
    project_path.touch()
    result = runner.run("--project", str(project_path))
    assert result.exit_code == golden_special.out["exit_code"]
    assert result.stdout == golden_special.out["output"]


def test_version(
    runner: Runner,
    httpx_mock: HTTPXMock,
) -> None:
    httpx_mock.add_response(
        url="http://localhost:5000/version",
        method="GET",
        json={"version": "1.2.3"},
    )

    result = runner.run("--version")

    assert result.exit_code == 0
    assert result.stdout == f"taas-client: {__version__}\ntaas server: 1.2.3\n"
